/*
  Copyright (c) 2020 Yuchen He <yhe306@gatech.edu> and Luis Alvarez <lalvarez@ulpgc.es>

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU Affero General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
  details.

  You should have received a copy of the GNU Affero General Public License along
  with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \file mainSkeleton2D.cpp
 * \brief main function to run the flux-ordered thinning algorithm
 * \author Yuchen He and Luis Alvarez
 */

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <cstring>
#include <string.h>
#include <time.h>

#include "skeleton.h"

using namespace std;


int main(int argc, char *argv[]){

  clock_t c_start = clock();
  /* ---------------------------------- INPUT PREPARATION ---------------------------------- */
  /// CHECK THE NUMBER OF PARAMETERS
  if(argc<9||argc>10){
    printf("argc=%d\n",argc);
    return -1;
  }

  /// READ THE IMAGE AND ITS NAME
  imageRGB I(argv[1]);
  if(I.size()==0){
    printf("problems reading image %s \n",argv[1]);
  }
  char name0[300],name[200];
  strncpy(name0,argv[1],strlen(argv[1])-4); /* extract the image name root */
  name0[strlen(argv[1])-4] = '\0';   /* null character manually added */


  /// READ THE THRESHOLD PARAMETER: GAMMA
  /* check the input */
  int num_punct = 0;
  for(size_t i=0; i<strlen(argv[3]);i++){
    if(!isdigit((argv[3])[i])){
      if((argv[3])[i]=='.' && num_punct==0){
        num_punct += 1;
        continue;
      }
      printf("The third parameter is a threshold of float type.\n");
      return -1;
    }
  }
  float gamma = atof(argv[3]);
  if(gamma<=0){
    throw("Gamma should be a positive real number.\n");
    return -1;
  }

  /// CONVERT RGB IMAGE TO GRAYSCALE
  image2D<float> L(I.width(),I.height());
  for(int k=0;k<I.size();k++){
    L[k]= (I[k].R_+I[k].G_+I[k].B_)/3.0;
  }

  /// INVERT THE BACKGROUND/SHAPE IF NECESSARY SO THAT THE SILHOUETTE IS IN BLACK COLOR
  int shape_color= atoi(argv[2]);
  if(shape_color==1){ /* if shape is white, invert*/
    for(int k=0;k<I.size();k++){
      L[k]= 1.0-L[k];
      I[k].R_ = 255.0 - I[k].R_;
      I[k].G_ = 255.0 - I[k].G_;
      I[k].B_ = 255.0 - I[k].B_;}
    }


  /// RENORMALIZE AND TRANSFORM TO BINARY SHAPE
  float maxL = L.max();
  float minL = L.min();
  for(int k=0;k<I.size();k++){
    L[k]= (L[k]-minL)/(maxL-minL);
    L[k]= L[k]>0.5?1.0:0.0;
  }

  image2D<float> L0=L; /* store L to compute similarity test */


  /// RECORD ORIGINAL SHAPE PIXELS
  int shape_pixel_num = 0;
  for(int k = 0; k < L.size(); k++){
    if(L[k]==0.0){
      shape_pixel_num += 1;
    }
  }

  /* ----------------------------------------------------------------------------------------- */
  clock_t c_end = clock();
  float time1 = 1000.0 * (c_end-c_start) / CLOCKS_PER_SEC;




  /* --------------------------- FLUX-ORDERD THINNING ALGOIRITHM ----------------------------- */
  c_start = clock();
  /* ---------------------------------------- PART I  ---------------------------------------- */
  /// COMPUTE THE DISTANCE FUNCTION INSIDE THE SILHOUETTE
  /* distance outside the sihouette is set to -1.0. */
  image2D<float> D;
  vector<point2d> contour;
  string distance_method;



  if(argc==10){
     distance_method = argv[9];
    if(distance_method.compare("brute_force") == 0){
      D=distance_function2D_contour(L,contour);
    }
    else if (distance_method.compare("F_H") == 0)
    {
      D=distance_function2D_F_H(L, contour);
    }
    else{
      distance_method = "fast_sweeping";
      D=distance_function2D_fast_sweeping(L, contour);
    }
  }else{
    distance_method = "fast_sweeping";
    D=distance_function2D_fast_sweeping(L, contour);
  }

  ofstream distance_data;
  distance_data.open("distance.dat");
  size_t im_w = D.width();
  size_t im_h = D.height();
  for (int i = 0; i < im_w; i++)
  {
    for (int j = 0; j < im_h; j++)
    {
      distance_data << D[i + j*im_w] << " ";
    }
    distance_data << endl;
  }
  distance_data.close();



  imageRGB image_to_write = D.convert2RGB(D.min(),D.max()); /* save the distance transform to disk*/

  strncpy(name,argv[4],strlen(argv[4]));
  name[strlen(argv[4])] = '\0';
  image_to_write.write(name);
  printf("Generate distance image... SUCCESS.\n");

  /// COMPUTE THE GRADIENT OF THE DISTANCE FUNCTION
  image2D<float> Dx,Dy;
  D.gradient(Dx,Dy);

  /// FLUX COMPUTATION USING IJCV PAPER "Hamilton-Jacobi Skeletons"
  image2D<float> F=flux(Dx,Dy);
  image_to_write=F.convert2RGB(F.min(),F.max()); /* save the average flux to disk*/
  strncpy(name,argv[5],strlen(argv[5]));
  name[strlen(argv[5])] = '\0';
  image_to_write.write(name);
  printf("Generate flux image... SUCCESS.\n");
  /* ----------------------------------------------------------------------------------------- */
  c_end = clock();
  float time2 = 1000.0 * (c_end-c_start) / CLOCKS_PER_SEC;

  c_start = clock();
  /* --------------------------------------- PART II  ---------------------------------------- */
  /// HOMOTOPY PRESERVING THINNING
  float flux_threshold = F.min()/gamma;

  image2D<float> original_shape(L);
  vector<pair<point2d,float>> medial_axis = homotopy_thinning(L, D, F, contour, flux_threshold);
  printf("Homotopy preserving thinning... SUCCESS.\n");

  /* ----------------------------------------------------------------------------------------- */
  c_end = clock();
  float time3 = 1000.0 * (c_end-c_start) / CLOCKS_PER_SEC;
  /* ----------------------- END OF FLUX-ORDERD THINNING ALGOIRITHM -------------------------- */




  /* --------------------------------- OUTPUT PREPARATION ------------------------------------ */
  /// CONTOUR SHAPE
  for(int i =0; i<F.size(); i++){
    image_to_write[i] = color((unsigned char) 255,(unsigned char) 255,(unsigned char) 255);
  }
  for(size_t i=0; i<contour.size();i++){
    int m = contour[i].x+contour[i].y*I.width();
    image_to_write[m] = color((unsigned char) 0,(unsigned char) 0,(unsigned char) 0);
  }
  /// HIGHLIGHT THE SKELETON POINTS USING RED
  for(int k=0;k<L.size();k++){
    if(L[k]==0.0){
      image_to_write[k]=color((unsigned char) 0,(unsigned char) 0,(unsigned char) 255);
    }
  }
  strncpy(name,argv[6],strlen(argv[6]));
  name[strlen(argv[6])] = '\0';
  image_to_write.write(name);
  printf("Generate skeleton image... SUCCESS.\n");


/// RECONSTRUCT SHAPE FROM SKELETON
  c_start = clock();
  printf("Reconstructing shape from skeleton...");
  image2D<float> canvas(L.width(),L.height(),1);


  float epsilon = atof(argv[8]); /* read in the dilation parameter epsilon */
  if(epsilon<0){
    throw("epsilon should be a non-negative real number.\n");
    return -1;
  }
  /// RECONSTRUCTION USING THE UNION OF DISKS ALONG THE MEDIAL AXIS
  shape_from_skeleton(canvas, medial_axis,epsilon);
  c_end = clock();
  float time4 = 1000.0 * (c_end-c_start) / CLOCKS_PER_SEC;
  /// store canvas to compute similarity test
  image2D<float> canvas0=canvas;
  if(shape_color==1){ /* recover to the orginal shape color if necessary */
    for(int m = 0; m < canvas.size(); m++){
      canvas[m] = 1.0-canvas[m];
    }
  }
  image_to_write=canvas.convert2RGB(canvas.min(),canvas.max());
  strncpy(name,argv[7],strlen(argv[7]));
  name[strlen(argv[7])] = '\0';
  image_to_write.write(name);
  printf("SUCCESS.\n");


  /* ----------------------------------------------------------------------------------------- */




  /* ---------------------------------- SUMMARY REPORT  -------------------------------------- */
  printf("********* SUMMARY *********\n");
  printf("Image dimension: %d, %d \n", I.width(), I.height());
  printf("The distance computation is based on: %s \n", distance_method.c_str());
  printf("Largest distance from inner point to the boundary: %lf \n", D.max());
  printf("Minimal flux: %lf, and Maximal flux: %lf \n",F.min(),F.max());
  printf("Threshold tau: %lf (gamma = %lf) \n", flux_threshold, gamma);
  printf("Total length of skeleton: %ld \n", medial_axis.size());
  printf("*** RECONSTRUCTION ***\n");
  printf("Dilation coefficient: %lf \n", epsilon);
  printf("Jaccard index: %lf \n", L0.Jacc(canvas0));
  printf("DSC similarity coefficient: %lf \n", L0.DSC(canvas0)); 
  printf("bias estimator Bpn : %lf \n", L0.Bpn(canvas0)); 
  printf("Total computation time: %lf ms \n", time1+time2+time3+time4);
  printf("--- Preparation: %lf ms \n", time1);
  printf("--- HJS: Part I: %lf ms \n", time2);
  printf("--- HJS: Part II: %lf ms \n", time3);
  printf("--- Reconstruct: %lf ms \n", time4);
  printf("***************************\n");
  /* ----------------------------------------------------------------------------------------- */

}
