/*
  Copyright (c) 2020 Yuchen He <yhe306@gatech.edu> and Luis Alvarez <lalvarez@ulpgc.es>

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU Affero General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
  details.

  You should have received a copy of the GNU Affero General Public License along
  with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \file skeleton.cpp
 * \brief definition of flux
 * \author Yuchen He and Luis Alvarez 
 */

#include "skeleton.h"



//////////////////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Compute the average outward flux of the gradient of distance map
*/
/// FLUX COMPUTATION
image2D<float> flux(image2D<float> &Ix, image2D<float> &Iy){
  /// COMPUTATION OF THE NEIGHBOR INDEX
  vector<int> N=Ix.Neighborhood();

  /// COMPUTATION OF NEIGHBOR NORMALS
  vector<float> Nx,Ny;
  Ix.NeighborhoodNormal(Nx,Ny);

  ///COMPUTATION OF THE FLUX
  image2D<float> Flux(Ix.width(),Ix.height(),0.);
  for(int y=1;y<Iy.height()-1;y++){
    int k0=y*Ix.width();
    for(int x=1;x<Ix.width()-1;x++){
      int k=k0+x;
      for(size_t m=0;m<N.size();m++){
        Flux[k]+=Ix[k+N[m]]*Nx[m]+Iy[k+N[m]]*Ny[m];
      }
      Flux[k]/=8.;
    }
  }

  return Flux;

}


/////////////////////////////////////////////////////////////////////////////////////////////
/**
  The following code is provided by an anonymous reviewer for a faster distance transform 
**/

// compute the slope of f between p and q
static float obtain_slope(float *f, int q, int p)
{
  if (!isfinite(f[q]) && !isfinite(f[p])) return 0;
  return ((f[q] + q*q) - (f[p] + p*p)) / (2*q - 2*p);
}

// compute the squared distances to a binary 1-D image
static void squared_distances_1d(
    float *d,  // output distance
    float *f,  // input data
    int n      // length
    )
{
  float z[n+1];  // temporary array
  int v[n];      // index table
  int k = 0;     // last index

  v[0] = 0;
  z[0] = -INFINITY;
  z[1] =  INFINITY;
  for (int q = 1; q < n; q++) {
    float s = obtain_slope(f, q, v[k]);
    while (s < z[k])
      s = obtain_slope(f, q, v[--k]);
    k++;
    v[k] = q;
    z[k] = s;
    z[k+1] = INFINITY;
  }

  k = 0;
  for (int q = 0; q < n; q++) {
    while (z[k+1] < q)
      k++;
    d[q] = (q - v[k]) * (q - v[k]) + f[v[k]];
  }
}

// compute the squared distances to 2-D image
static void squared_distances_2d(float *f, int w, int h)
{
  // columns
  for (int i = 0; i < w; i++)
  {
    float t[h], d[h];
    for (int j = 0; j < h; j++) t[j] = f[j*w+i];
    squared_distances_1d(d, t, h);
    for (int j = 0; j < h; j++) f[j*w+i] = d[j];
  }

  // rows
  for (int j = 0; j < h; j++)
  {
    float t[w], d[w];
    for (int i = 0; i < w; i++) t[i] = f[j*w+i];
    squared_distances_1d(d, t, w);
    for (int i = 0; i < w; i++) f[j*w+i] = d[i];
  }
}

void squared_euclidean_distance_to_nonzeros(
    float *x,   // input/output: (binary image/distance to nonzeros)
    int w,      // width
    int h       // height
    )
{
  for (int i = 0; i < w*h; i++)
    x[i] = x[i] > 0 ? 0: INFINITY;
  squared_distances_2d(x, w, h);
}

void build_positive_distance(float *d, float *m, int w, int h)
{
  for (int i = 0; i < w*h; i++)
    d[i] = m[i] > 0 ? 0 : INFINITY;
  squared_distances_2d(d, w, h);
  for (int i = 0; i < w*h; i++)
    d[i] = sqrt(d[i]);
}

